package common.tags;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.Map;
import java.util.Map.Entry;

import common.collect.Maps;

import java.util.Set;

public class TagObject extends Tag {
	private Map<String, Tag> tags = Maps.<String, Tag>newHashMap();

	void write(DataOutput output) throws IOException {
		for(String key : this.tags.keySet()) {
			Tag tag = this.tags.get(key);
			output.writeByte(tag.getId());
			if(tag.getId() != 0) {
				output.writeUTF(key);
				tag.write(output);
			}
		}
		output.writeByte(0);
	}

	void read(DataInput input, int depth, SizeTracker tracker) throws IOException {
		tracker.read(48);
		if(depth > 512)
			throw new RuntimeException("Objekt ist zu komplex, die Tiefe ist größer als 512");
		this.tags.clear();
		byte id;
		while((id = input.readByte()) != 0) {
			String key = input.readUTF();
			tracker.read(28 + 2 * key.length());
			Tag tag = Tag.createNewByType(id);
			tag.read(input, depth + 1, tracker);
			if(this.tags.put(key, tag) != null)
				tracker.read(36);
		}
	}

	public Set<String> getKeySet() {
		return this.tags.keySet();
	}

	protected byte getId() {
		return (byte)10;
	}

	public void set(String key, Tag value) {
		this.tags.put(key, value);
	}

	public void setTag(String key, TagObject value) {
		this.tags.put(key, value);
	}

	public void setFloatList(String key, TagFloatList value) {
		this.tags.put(key, value);
	}

	public void setDoubleList(String key, TagDoubleList value) {
		this.tags.put(key, value);
	}

	public void setStringList(String key, TagStringList value) {
		this.tags.put(key, value);
	}

	public void setIntArrayList(String key, TagIntArrayList value) {
		this.tags.put(key, value);
	}

	public void setTagList(String key, TagObjectList value) {
		this.tags.put(key, value);
	}

	public void setByte(String key, byte value) {
		this.tags.put(key, new TagByte(value));
	}

	public void setShort(String key, short value) {
		this.tags.put(key, new TagShort(value));
	}

	public void setInt(String key, int value) {
		this.tags.put(key, new TagInt(value));
	}

	public void setLong(String key, long value) {
		this.tags.put(key, new TagLong(value));
	}

	public void setFloat(String key, float value) {
		this.tags.put(key, new TagFloat(value));
	}

	public void setDouble(String key, double value) {
		this.tags.put(key, new TagDouble(value));
	}

	public void setString(String key, String value) {
		this.tags.put(key, new TagString(value));
	}

	public void setByteArray(String key, byte[] value) {
		this.tags.put(key, new TagByteArray(value));
	}

	public void setIntArray(String key, int[] value) {
		this.tags.put(key, new TagIntArray(value));
	}

	public void setBool(String key, boolean value) {
		this.setByte(key, (byte)(value ? 1 : 0));
	}

	public Tag get(String key) {
		return this.tags.get(key);
	}

	private boolean has(String key, int type) {
		Tag tag = this.tags.get(key);
		return tag != null && tag.getId() == type;
	}

	public boolean hasBool(String key) {
		return this.hasByte(key);
	}

	public boolean hasByte(String key) {
		return this.has(key, 1);
	}

	public boolean hasShort(String key) {
		return this.has(key, 2);
	}

	public boolean hasInt(String key) {
		return this.has(key, 3);
	}

	public boolean hasLong(String key) {
		return this.has(key, 4);
	}

	public boolean hasFloat(String key) {
		return this.has(key, 5);
	}

	public boolean hasDouble(String key) {
		return this.has(key, 6);
	}

	public boolean hasString(String key) {
		return this.has(key, 8);
	}

	public boolean hasByteArray(String key) {
		return this.has(key, 7);
	}

	public boolean hasIntArray(String key) {
		return this.has(key, 11);
	}

	public boolean hasTag(String key) {
		return this.has(key, 10);
	}

	public boolean hasFloatList(String key) {
		return this.has(key, 13);
	}

	public boolean hasDoubleList(String key) {
		return this.has(key, 14);
	}

	public boolean hasStringList(String key) {
		return this.has(key, 9);
	}

	public boolean hasIntArrayList(String key) {
		return this.has(key, 15);
	}

	public boolean hasTagList(String key) {
		return this.has(key, 12);
	}

	public byte getByte(String key) {
		return !this.has(key, 1) ? 0 : ((TagByte)this.tags.get(key)).getByte();
	}

	public short getShort(String key) {
		return !this.has(key, 2) ? 0 : ((TagShort)this.tags.get(key)).getShort();
	}

	public int getInt(String key) {
		return !this.has(key, 3) ? 0 : ((TagInt)this.tags.get(key)).getInt();
	}

	public long getLong(String key) {
		return !this.has(key, 4) ? 0L : ((TagLong)this.tags.get(key)).getLong();
	}

	public float getFloat(String key) {
		return !this.has(key, 5) ? 0.0F : ((TagFloat)this.tags.get(key)).getFloat();
	}

	public double getDouble(String key) {
		return !this.has(key, 6) ? 0.0D : ((TagDouble)this.tags.get(key)).getDouble();
	}

	public String getString(String key) {
		return !this.has(key, 8) ? "" : ((TagString)this.tags.get(key)).getString();
	}

	public byte[] getByteArray(String key) {
		return !this.has(key, 7) ? new byte[0] : ((TagByteArray)this.tags.get(key)).getByteArray();
	}

	public int[] getIntArray(String key) {
		return !this.has(key, 11) ? new int[0] : ((TagIntArray)this.tags.get(key)).getIntArray();
	}

	public TagObject getTag(String key) {
		return !this.has(key, 10) ? new TagObject() : (TagObject)this.tags.get(key);
	}

	public TagFloatList getFloatList(String key) {
		return !this.has(key, 13) ? new TagFloatList() : (TagFloatList)this.tags.get(key);
	}

	public TagDoubleList getDoubleList(String key) {
		return !this.has(key, 14) ? new TagDoubleList() : (TagDoubleList)this.tags.get(key);
	}

	public TagStringList getStringList(String key) {
		return !this.has(key, 9) ? new TagStringList() : (TagStringList)this.tags.get(key);
	}

	public TagIntArrayList getIntArrayList(String key) {
		return !this.has(key, 15) ? new TagIntArrayList() : (TagIntArrayList)this.tags.get(key);
	}

	public TagObjectList getTagList(String key) {
		return !this.has(key, 12) ? new TagObjectList() : (TagObjectList)this.tags.get(key);
	}

	public boolean getBool(String key) {
		return this.getByte(key) != 0;
	}

	public void remove(String key) {
		this.tags.remove(key);
	}

	public String toString() {
		StringBuilder sb = new StringBuilder("{");
		for(Entry<String, Tag> entry : this.tags.entrySet()) {
			if(sb.length() != 1)
				sb.append(',');
			sb.append(entry.getKey()).append(':').append(entry.getValue());
		}
		return sb.append('}').toString();
	}

	public boolean isEmpty() {
		return this.tags.isEmpty();
	}

	public Tag copy() {
		TagObject tag = new TagObject();
		for(String s : this.tags.keySet()) {
			tag.set(s, this.tags.get(s).copy());
		}
		return tag;
	}

	public boolean equals(Object other) {
		return super.equals(other) && this.tags.entrySet().equals(((TagObject)other).tags.entrySet());
	}

	public int hashCode() {
		return super.hashCode() ^ this.tags.hashCode();
	}

	/**
	 * Merges this NBTTagCompound with the given compound. Any sub-compounds are
	 * merged using the same methods, other types of tags are overwritten from the
	 * given compound.
	 */
	public void merge(TagObject other) {
		for(String key : other.tags.keySet()) {
			Tag tag = other.tags.get(key);
			if(tag.getId() == 10) {
				if(this.has(key, 10)) {
					TagObject comp = this.getTag(key);
					comp.merge((TagObject)tag);
				}
				else {
					this.set(key, tag.copy());
				}
			}
			else {
				this.set(key, tag.copy());
			}
		}
	}
}
